""" simple cache based on shelve
"""

import shelve
import time


def load_cache(cache_file):
    cache = shelve.open(cache_file)
    return cache


def read_key(cache, key, ttl):
    """ Reads value from cache, if doesnt exist or is older than ttl, raises KeyError
    """
    bibtex, timestamp = cache[key]

    if (timestamp + ttl) < time.time():
        raise KeyError("Cached entry is too old")
    else:
        return bibtex


def save_key(cache, key, value):
    cache[key] = (value, time.time())



""" Main file
"""

import yaml
import re

import query_ads
import adsbibtex_cache
import adsbibtex_exceptions


def run_adsbibtex(config_file):
    config_document = load_config_file(config_file)
    yaml_front_matter, bibcode_lines = parse_config_file(config_document)

    config = parse_yaml_front_matter(yaml_front_matter)

    bibcode_list = parse_bibcode_lines(bibcode_lines)

    cache = adsbibtex_cache.load_cache(config['cache_file'])
    cache_ttl = config['cache_ttl'] * 3600  # to seconds

    for i, bibcode_entry in enumerate(bibcode_list):
        bibcode = bibcode_entry['bibcode']
        try:
            bibtex = adsbibtex_cache.read_key(cache, bibcode, cache_ttl)
            print '{} successfully fetched from cache'.format(bibcode)
        except KeyError:  # bibcode not cached or old
            try:
                bibtex = query_ads.bibcode_to_bibtex(bibcode)
                adsbibtex_cache.save_key(cache, bibcode, bibtex)
                print '{} successfully fetched from ADS'.format(bibcode)
            except adsbibtex_exceptions.ADSBibtexBibcodeNotFound:
                print '{} not found on ADS, skipping'.format(bibcode)
                continue

        bibcode_entry['bibtex'] = bibtex

    cache.close()

    bibtex_ouput = generate_bibtex_output(bibcode_list)

    save_bibtex_output(bibtex_ouput, out_path=config['bibtex_file'])


def load_config_file(config_file):
    """ Loads the given file into a list of lines

    :param config_file: file name of the config file
    :type config_file: str

    :return: config file as a list (one item per line) as returned by open().readlines()
    """
    with open(config_file, 'r') as f:
        config_document = f.readlines()
    return config_document


def parse_config_file(config_document):
    """ Parses the config file into a yaml front matter and a list of bibcode lines

    :param config_file: config file as a list (one item per line) as returned by open().readlines()
    :type config_file: [str, str, ...]

    :return: yaml_front_matter, bibcode_lines
    :rtype: str, list
    """

    yaml_front_matter = []
    i = 0
    yaml_front_matter_found = False
    for i, line in enumerate(config_document):
        if line.startswith('---'):
            yaml_front_matter_found = True
            break
        else:
            yaml_front_matter.append(line)

    if not yaml_front_matter_found:
        raise adsbibtex_exceptions.ADSBibtexConfigError("YAML Front matter not found")

    yaml_front_matter = ''.join(yaml_front_matter)

    bibcode_lines = config_document[i + 1:]

    return yaml_front_matter, bibcode_lines


def parse_yaml_front_matter(yaml_front_matter):
    """ Parses the YAML front matter and checks for essential config params

    :param yaml_front_matter:
    :return: config
    :rtype: dict
    """

    config = yaml.load(yaml_front_matter)

    # TODO check essential config params / set defaults

    return config


def parse_bibcode_lines(bibcode_lines):
    """ Cleans up the bibcode line list, removing comments and returning a list of
    {"cite_name": cite_name, "bibcode": bibcode} pairs

    :param bibcode_lines: lists of bibcode lines
    :return:
    """

    bibcode_list = []

    for bibcode_line in bibcode_lines:
        try:
            cite_name, bibcode = parse_bibcode_line(bibcode_line)
            bibcode_list.append({'cite_name': cite_name, 'bibcode': bibcode})
        except TypeError:  # None was returned
            pass

    return bibcode_list


def parse_bibcode_line(bibcode_line):
    """ Parses a single bibcode line into cite_name, bibcode

    :param bibcode_line: a line containing bibcode [cite_name] [# comment]
    :return: cite_name, bibcode
    """
    # TODO (ryan) think about replacing with regex
    if not is_comment(bibcode_line):  # line is just a comment
        bibcode_line = bibcode_line.replace('\t', ' ')
        bibcode_line = bibcode_line.replace('\n', '')
        bibcode_line = ' '.join(bibcode_line.split())
        # replace multiple whitespace with single space

        bibcode_split = bibcode_line.split(' ', 2)
        num_splits = len(bibcode_split)
        if num_splits == 1:
            bibcode = cite_name = bibcode_split[0]
        elif num_splits > 1:
            bibcode, cite_name = bibcode_split[:2]
            if is_comment(cite_name):
                cite_name = bibcode
        else:
            return None

        if bibcode and cite_name:
            return cite_name, bibcode
        else:
            return None
    else:
        return None


def is_comment(s):
    """ Checks if a string looks like a comment line (first non whitespace char is #)
    :param s: string to check
    :return: True/False
    """
    if re.match(r'(\s+|)#(.*?)', s):
        return True
    else:
        return False


def replace_bibtex_cite_name(bibtex, current_name, new_name):
    """replaces the cite_name in a bibtex file with something else

    :param: string of bibtex to do the replacing on
    :param current_name: current cite name in the bibtex
    :param new_name: name to replace it with
    """

    new_bibtex = bibtex.replace(current_name, new_name, 1)

    return new_bibtex


def generate_bibtex_output(bibcode_list):
        output = []
        for bibcode_item in bibcode_list:
            try:
                bibtex = bibcode_item['bibtex']
            except KeyError:  # Bibcode not found
                continue

            bibcode = bibcode_item['bibcode']
            cite_name = bibcode_item['cite_name']
            new_bibtex = replace_bibtex_cite_name(bibtex, bibcode, cite_name)

            output.append(new_bibtex)

        return ''.join(output)


def save_bibtex_output(bibtex_output, out_path):
    with open(out_path, 'w') as f:
        f.write(bibtex_output)



""" queries ads using https://github.com/adsabs/adsabs-dev-api
"""

import ads.exceptions
from ads.export import ExportQuery


def bibcode_to_bibtex(bibcode):
    """Queries ads for the bibtex"""

    try:
        bibtex = ExportQuery(bibcodes=bibcode, format="bibtex").execute()
        return bibtex
    except ads.exceptions.APIResponseError as e:
        raise e  # TODO (ryan) provide rate limit info


from setuptools import setup
import codecs
import os
import re

here = os.path.abspath(os.path.dirname(__file__))


# Read the version number from a source file.
def find_version(*file_paths):
    # Open in Latin-1 so that we avoid encoding errors.
    # Use codecs.open for Python 2 compatibility
    with codecs.open(os.path.join(here, *file_paths), 'r', 'latin1') as f:
        version_file = f.read()

    # The version line must have the form
    # __version__ = 'ver'
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              version_file, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Unable to find version string.")


# Get the long description from the relevant file
with codecs.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

install_requires = ['nose>=1.0', 'hypothesis', 'docopt', 'ads']

setup(
    name="ADSBibTeX",
    version=find_version('adsbibtex', '__init__.py'),
    description="ADSBibTeX",
    long_description=long_description,
    url='https://github.com/ryanvarley/adsbibtex',
    author='Ryan Varley',
    author_email='ryan@ryanvarley.uk',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Education',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Operating System :: MacOS',
        'Programming Language :: Python :: 2.7',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],

    entry_points={
        'console_scripts': [
            'adsbibtex = adsbibtex.bibcode_to_bibtex:run',
        ],
    },

    # What does your project relate to?
    # keywords='sample setuptools development',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages.
    packages=['adsbibtex'],

    # List run-time dependencies here.  These will be installed by pip when your
    # project is installed.
    install_requires=install_requires,

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    include_package_data=True,
    zip_safe=False,
    test_suite='nose.collector'
)



#!/usr/bin/env python
""" Converts a ADSBibTeX config file to a BibTeX file

Usage:
  adsbibtex [<config_file>]

Options:
    config    config file location
"""

import docopt
import adsbibtex


def run():
    arguments = docopt.docopt(__doc__)
    config_path = arguments['<config_file>']

    if config_path is None:  # use default
        config_path = 'config.adsbib'

    adsbibtex.run_adsbibtex(config_path)


if __name__ == '__main__':
    run()
